#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "symtab.h"
#include "tfscan.h"

#define BUFSIZE    256     /* file read buffer size */

static NIMAP  *nimap_item = NULL;       /* for session field and item field */
static TFSCAN *tfscan = NULL;           /* table file scanner  */
FILE  *tablefile, *columnfile, *sessionfile, *itemfile, *clusterfile;

void error(char *errmsg)
{
  fprintf(stderr, "%s\n", errmsg);
  exit(1);
}

int map_one_field(NIMAP *nimap, char *buf)
{
  int *item;
  
  item = nim_add(nimap, buf, sizeof(int));
  if (item == NULL) return -1;
  if (item == EXISTS) { 
    item = (int *)nim_byname(nimap, buf);
  }
  assert(item);
  return(*item);
}

int ReadRecord(int curline, int colcount, char **colname)
{
  struct {
    unsigned long sid, iid;
  } data;
  char  buf[BUFSIZE];
  int   d;
  int   curfield = 0;
  char  tmpstr[256];
  int   iid = -1;

  assert( nimap_item && tfscan && (colcount>0));

  while (1) {
    d = tfs_getfld(tfscan, tablefile, buf, BUFSIZE);
    if (d <= TFS_EOF ) {  /* error(-1) or EOF(0) */
      if (d < 0)  error("Error reading table file."); 
      else break; 
    }
    if (buf[0] == '\0') { /* nothing read */
      if (d >= TFS_FLD) error("Item expected.");
      printf("Empty line: %d", curline+1);
      break;
    }
    if (curfield == 0) { /* first field: cluster result */
      fprintf(clusterfile, "%s\n", buf);
    }
    else if (curfield <= colcount) { /* feature values */
      if (buf[0] != '?') {
	sprintf(tmpstr, "%s*%s", colname[curfield-1], buf);
	iid = map_one_field(nimap_item, tmpstr);
	if (iid < 0) error("ItemName/ID mapping failed.");
	data.sid = curline;
	data.iid = iid;
	fwrite((const void *)&data, 4, 2, sessionfile);
      } else {
	/*	printf("%s\n", buf); */
      }
    }
    curfield++;
    if (d == TFS_REC)  break;  
  }
  if ((curfield != 0) && (curfield != (colcount+1))) {
    fprintf(stderr, "Line %d[%s]: %d fields instead of %d.\n", curline, 
	    buf, curfield, colcount+1);
    return TFS_REC;
  } 

  return d;
}

int main(int argc, char *argv[])
{
  char  tablefilename[256], columnfilename[256];  /* input file */
  char  sessionfilename[256], itemfilename[256], clusterfilename[256];  /* output file */
  int   colcount = 0;
  char  **colname;
  char  tmpstr[256];
  int   i, d;
  int   curline, maxitemid;
  char  fldseps[10]=",";
  
  printf("\ndata2session: transfer the table file to session file.\n");
  if (argc != 4) {
    printf("Usage: %s tablefile columnfile outputfile\n", argv[0]);
    printf("tablefile --the first column is the clustering result, the other colums are feature values.\n"); 
    printf("columnfile -- the feature name extracting from the table file, the first line is the feature count.\n");
    printf("outputfile -- session file name.\n");
    error("\nWrong number of arguments.\n");
    
    /* try:  data2session mush.data column mush */
  }
  /* table file is from www.ics.uci.edu/mlearn, the first column is the result of 
     cluster, the other colums are feature values */
  strcpy(tablefilename, argv[1]);
  /* the column name extract from the table file, the first line is the feature count */
  strcpy(columnfilename, argv[2]);
  /* the result file store the created session (binary file) */
  sprintf(sessionfilename, "%s.session",argv[3]);
  /* store the index from id to ID */
  sprintf(itemfilename, "%s.item",argv[3]);
  /* store the clusters extracted from tablefile */
  sprintf(clusterfilename, "%s.cluster",argv[3]);

  tablefile = fopen(tablefilename, "r");
  if (!tablefile)  error("Cannot open tablefile to read.");

  columnfile = fopen(columnfilename, "r");
  if (!columnfile)  error("Cannot open columnfile to read.");

  sessionfile = fopen(sessionfilename, "wb");
  if (!sessionfile)  error("Cannot open sessionfile to write.");

  itemfile = fopen(itemfilename, "w");
  if (!itemfile)  error("Cannot open itemfile to write.");

  clusterfile = fopen(clusterfilename, "w");
  if (!clusterfile)  error("Cannot open clusterfile to write.");

  tfscan = tfs_create();
  if (!tfscan)  error("Not enough memory!");
  tfs_chars(tfscan, TFS_FLDSEP, fldseps);
  nimap_item = nim_create(0, 0, (HASHFN*)0, (SYMFN*)0);
  if (!nimap_item)  error("Not enough memory!");

  /* read the column file */
  fscanf(columnfile, "%d\n", &colcount);
  if (0 < colcount) {
    colname = (char **)calloc(colcount, sizeof(char *));
    if (!colname)  error("Not enough memory!");
  } else {
    error("The column count is no more than 0.");
  }
  i = 0;
  while (fscanf(columnfile, "%s\n", tmpstr) && (i < colcount)) {
    colname[i] = (char *)malloc(strlen(tmpstr) + 1);
    if (!colname[i])  error("Not enough memory!");
    strcpy(colname[i], tmpstr);
    i++;
  }
  /* read the table file */
  curline = 0;  maxitemid = 0;

  printf("Reading the data file ....\n");
  /* leave 4 bytes space for the TotalSessions field */  
  fwrite((const void *)&curline, 4, 1, sessionfile);
  /* leave 4 bytes space for the MaxItemID field */  
  fwrite((const void *)&maxitemid, 4, 1, sessionfile);

  while (1) {
    d = ReadRecord(curline, colcount, colname);
    if (d <= TFS_EOF) break;
    curline++;
  }
  rewind(sessionfile);
  /* write the TotalSession field at the beginning*/  
  fwrite((const void *)&curline, 4, 1, sessionfile); 
  /* write the TotalSession field at the beginning*/
  maxitemid = nim_cnt(nimap_item);
  fwrite((const void *)&maxitemid, 4, 1, sessionfile); 

  d = nim_cnt(nimap_item);
  for (i = 0; i < d; i++)
    fprintf(itemfile, "%s\n", nim_name(nim_byid(nimap_item, i)));

  tfs_delete(tfscan);
  nim_delete(nimap_item);
  
  fclose(tablefile);
  fclose(columnfile);
  fclose(sessionfile);
  fclose(itemfile);
  fclose(clusterfile);

  printf("Files %s, %s, %s written.\n\n", sessionfilename, itemfilename, clusterfilename);
  
  for(i = 0; i < colcount; i++)  free(colname[i]);
  free(colname);

  return 0;
}
